from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives
from eth_typing import HexAddress
from pydantic import BaseModel

ROUTER_ADDRESS = "0xBE6D8f0d05cC4be24d5167a3eF062215bE6D18a5"


class ExactInputSingleParams(BaseModel):
    token_in: Annotated[HexAddress, Name("tokenIn")]
    token_out: Annotated[HexAddress, Name("tokenOut")]
    tick_spacing: Annotated[primitives.int24, Name("tickSpacing")]
    recipient: HexAddress
    deadline: primitives.uint256
    amount_in: Annotated[primitives.uint256, Name("amountIn")]
    amount_out_minimum: Annotated[primitives.uint256, Name("amountOutMinimum")]
    sqrt_price_limit_x96: Annotated[primitives.uint160, Name("sqrtPriceLimitX96")]


class SlipstreamRouter(ProtocolBase):
    exact_input_single: Annotated[
        ContractFunc[ExactInputSingleParams, primitives.uint256],
        Name("exactInputSingle"),
    ] = METHOD



from typing import Annotated

from eth_rpc import ContractFunc, Event, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, Struct, primitives
from pydantic import BaseModel, Field

from .utils import Command


class NewBlockEventType(BaseModel):
    index: primitives.uint256
    hash: primitives.bytes32


NewBlockEvent = Event[NewBlockEventType](name="NewBlock")


class SimulacrumBlock(Struct):
    number: primitives.uint256
    timestamp: primitives.uint256
    commands: list[primitives.bytes32]
    command_count: primitives.uint256
    hash: primitives.bytes32
    mined: bool


class Blocks(ProtocolBase):
    mine_block: Annotated[
        ContractFunc[
            primitives.uint256,
            None,
        ],
        Name("mineBlock"),
    ] = METHOD

    get_command: Annotated[
        ContractFunc[primitives.bytes32, Command],
        Name("getCommand"),
    ] = METHOD

    get_nonce: Annotated[
        ContractFunc[
            Command,
            primitives.uint256,
        ],
        Name("getNonce"),
    ] = METHOD

    get_block_commands: Annotated[
        ContractFunc[
            primitives.uint256,
            list[Command],
        ],
        Name("getBlockCommands"),
    ] = METHOD

    default_gas: Annotated[
        ContractFunc[
            NoArgs,
            primitives.uint256,
        ],
        Name("defaultGas"),
    ] = METHOD

    pending_txs: Annotated[
        ContractFunc[
            NoArgs,
            primitives.uint256,
        ],
        Name("pendingTxs"),
    ] = METHOD

    get_block: Annotated[
        ContractFunc[
            primitives.uint256,
            SimulacrumBlock,
        ],
        Name("getBlock"),
    ] = METHOD

    current_block: Annotated[
        ContractFunc[
            NoArgs,
            primitives.uint256,
        ],
        Name("currentBlock"),
    ] = METHOD


class NonceArgs(BaseModel):
    owner_id: primitives.bytes32
    namespace: primitives.bytes32
    index: primitives.uint256 = Field(default=primitives.uint256(0))



from eth_rpc import ContractFunc
from eth_rpc.networks import Base
from eth_rpc.wallet import PrivateKeyWallet
from eth_typing import HexAddress, HexStr

from .executor import ExecuteRequest, SimulacrumExecutor


class Executor:
    def __init__(
        self,
        simulacrum_wallet_address: HexAddress,
        executor_wallet: PrivateKeyWallet,
        simulacrum_executor: SimulacrumExecutor | None = None,
    ):
        self.simulacrum_wallet_address = simulacrum_wallet_address
        self.executor_wallet = executor_wallet
        self.simulacrum_executor = simulacrum_executor
        self.executor = simulacrum_executor or SimulacrumExecutor[Base](
            address=self.simulacrum_wallet_address
        )

    async def broadcast(
        self,
        contract_func: ContractFunc,
        value: int = 0,
        operation: int = 0,
        tx_gas: int = 100_000,
        executor_value: int = 0,
    ) -> HexStr:
        response_tx = await self.executor.execute(
            ExecuteRequest(
                to=contract_func.address,
                value=value,
                data=contract_func.encode(),
                operation=operation,
                txGas=tx_gas,
            )
        ).execute(wallet=self.executor_wallet, value=executor_value)
        return response_tx



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, Struct, primitives


class TokenStruct(Struct):
    name: primitives.bytes32
    token_address: primitives.address


class TokenRegistry(ProtocolBase):
    lookup: ContractFunc[
        primitives.bytes32,
        primitives.address,
    ] = METHOD

    update_many: Annotated[
        ContractFunc[
            tuple[primitives.bytes32, list[TokenStruct]],
            None,
        ],
        Name("updateMany"),
    ] = METHOD



from typing import Annotated

from eth_rpc import ContractFunc
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typeshed.erc20 import ERC20


class WETH(ERC20):
    withdraw: Annotated[
        ContractFunc[primitives.uint256, NoArgs],
        Name("withdraw"),
    ] = METHOD
    deposit: Annotated[
        ContractFunc[NoArgs, NoArgs],
        Name("deposit"),
    ] = METHOD



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives


class AttestationOracle(ProtocolBase):
    verify: ContractFunc[
        tuple[
            Annotated[primitives.bytes32, Name("dataType")],
            Annotated[primitives.bytes, Name("sourceData")],
            Annotated[primitives.bytes, Name("verificationData")],
        ],
        Annotated[primitives.bool, Name("success")],
    ] = METHOD



from decimal import Decimal
from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typing import HexAddress
from pydantic import BaseModel, Field, PrivateAttr

from ..erc20 import ERC20

Q192 = Decimal(2**192)


class GlobalState(BaseModel):
    price: primitives.uint160
    tick: primitives.int24
    fee_zto: primitives.uint16 = Field(serialization_alias="feeZto")
    fee_0tz: primitives.uint16 = Field(serialization_alias="feeOtz")
    timepoint_index: primitives.uint16 = Field(serialization_alias="timepointIndex")
    community_fee_token0: primitives.uint8 = Field(
        serialization_alias="communityFeeToken0"
    )
    community_fee_token1: primitives.uint8 = Field(
        serialization_alias="communityFeeToken1"
    )
    unlocked: bool


class CamelotV3Pool(ProtocolBase):
    _tick_spacing: int | None = PrivateAttr(None)
    _token0: ERC20 | None = PrivateAttr(None)
    _token1: ERC20 | None = PrivateAttr(None)

    fee: ContractFunc[NoArgs, primitives.uint24] = METHOD
    global_state: Annotated[ContractFunc[NoArgs, GlobalState], Name("globalState")] = (
        METHOD
    )
    token0: ContractFunc[NoArgs, HexAddress] = METHOD
    token1: ContractFunc[NoArgs, HexAddress] = METHOD

    async def get_price(self, token0: bool = True):
        global_state = await self.global_state().get()
        _token0 = await self.token0().get()
        _token1 = await self.token1().get()
        token0_decimals = await ERC20(address=_token0).decimals().get()
        token1_decimals = await ERC20(address=_token1).decimals().get()
        return self.sqrt_price_x96_to_token_prices(
            global_state.price,
            token0_decimals,
            token1_decimals,
            token0,
        )

    @staticmethod
    def sqrt_price_x96_to_token_prices(
        sqrt_price_x96: int,
        token0_decimals: int,
        token1_decimals: int,
        token0: bool = True,
    ) -> Decimal:
        num = Decimal(sqrt_price_x96 * sqrt_price_x96)
        price1 = (
            (num / Q192)
            * (Decimal(10) ** token0_decimals)
            / (Decimal(10) ** token1_decimals)
        )
        price0 = Decimal(1) / price1

        return price0 if token0 else price1

    def __repr__(self):
        return f"<Pool address={self.address}>"



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives
from pydantic import BaseModel


class ExecuteRequest(BaseModel):
    to: primitives.address
    value: primitives.uint256
    data: primitives.bytes
    operation: primitives.uint8
    txGas: primitives.uint256


class ExecuteResponse(BaseModel):
    success: primitives.bool
    returnData: primitives.bytes


class SimulacrumExecutor(ProtocolBase):
    execute: Annotated[
        ContractFunc[
            ExecuteRequest,
            ExecuteResponse,
        ],
        Name("execute"),
    ] = METHOD



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives


class CommandTracker(ProtocolBase):
    errors: ContractFunc[primitives.bytes32, bytes] = METHOD

    get_error: Annotated[
        ContractFunc[primitives.bytes32, bytes],
        Name("getError"),
    ] = METHOD

    set_error: Annotated[
        ContractFunc[
            tuple[primitives.bytes32, primitives.bytes32, primitives.bytes32, bytes],
            None,
        ],
        Name("setError"),
    ] = METHOD

    set_success: Annotated[
        ContractFunc[
            tuple[primitives.bytes32, primitives.bytes32, primitives.bytes32, bytes],
            None,
        ],
        Name("setSuccess"),
    ] = METHOD

    success: ContractFunc[primitives.bytes32, bytes] = METHOD



# flake8: noqa

from .weighted_index import (
    AddLiquidityEvent,
    AddLiquidityType,
    BondEvent,
    BondType,
    DebondEvent,
    DebondType,
    RemoveLiquidityEvent,
    RemoveLiquidityType,
)



from eth_rpc import get_current_network
from eth_rpc.networks import Arbitrum, Ethereum
from eth_rpc.types import Network
from eth_typing import HexAddress, HexStr


class Tokens:
    class Ethereum:
        WETH = HexAddress(HexStr("0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2"))
        USDC = HexAddress(HexStr("0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48"))
        USDT = HexAddress(HexStr("0xdac17f958d2ee523a2206206994597c13d831ec7"))
        DAI = HexAddress(HexStr("0x6b175474e89094c44da98b954eedeac495271d0f"))

        main = [WETH, USDC, USDT, DAI]
        stables = [USDC, USDT, DAI]

    class Arbitrum:
        WETH = HexAddress(HexStr("0x82af49447d8a07e3bd95bd0d56f35241523fbab1"))
        USDC = HexAddress(HexStr("0xaf88d065e77c8cc2239327c5edb3a432268e5831"))
        USDT = HexAddress(HexStr("0xfd086bc7cd5c481dcc9c85ebe478a1c0b69fcbb9"))

        main = [WETH, USDC, USDT]
        stables = [USDC, USDT]

    @classmethod
    def for_network(cls, network: type[Network] = get_current_network()):
        if network.chain_id == Ethereum.chain_id:
            return cls.Ethereum
        if network.chain_id == Arbitrum.chain_id:
            return cls.Arbitrum


class Routers:
    class Ethereum:
        UniswapV2 = HexAddress(HexStr("0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D"))
        UniswapV3 = HexAddress(HexStr("0xe592427a0aece92de3edee1f18e0157c05861564"))

        podrouter = UniswapV2

    class Arbitrum:
        UniswapV2 = HexAddress(HexStr("0x4752ba5dbc23f44d87826276bf6fd6b1c372ad24"))
        UniswapV3 = HexAddress(HexStr("0xe592427a0aece92de3edee1f18e0157c05861564"))
        Camelot = HexAddress(HexStr("0xc873fEcbd354f5A56E00E710B90EF4201db2448d"))

        pod_router = Camelot

    class Linea:
        Lynex = HexAddress(HexStr("0x610D2f07b7EdC67565160F587F37636194C34E74"))

    @classmethod
    def for_network(cls, network: type[Network] = get_current_network()):
        if network.chain_id == Ethereum.chain_id:
            return cls.Ethereum
        if network.chain_id == Arbitrum.chain_id:
            return cls.Arbitrum


class Factories:
    class Ethereum:
        UniswapV2 = HexAddress(HexStr("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f"))
        UniswapV3 = HexAddress(HexStr("0x1F98431c8aD98523631AE4a59f267346ea31F984"))
        SushiSwap = HexAddress(HexStr("0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac"))
        ShibaSwap = HexAddress(HexStr("0x115934131916C8b277DD010Ee02de363c09d037c"))

        pod_factory = UniswapV2
        all_factories = [UniswapV2, UniswapV3]

    class Arbitrum:
        UniswapV2 = HexAddress(HexStr("0xf1D7CC64Fb4452F05c498126312eBE29f30Fbcf9"))
        UniswapV3 = HexAddress(HexStr("0x1F98431c8aD98523631AE4a59f267346ea31F984"))
        Camelot_V2 = HexAddress(HexStr("0x6EcCab422D763aC031210895C81787E87B43A652"))
        Camelot_V3 = HexAddress(HexStr("0x1a3c9B1d2F0529D97f2afC5136Cc23e58f1FD35B"))

        pod_factory = Camelot_V2
        all_factories = [UniswapV2, UniswapV3, Camelot_V2, Camelot_V3]

    class Linea:
        Lynex = HexAddress(HexStr("0xBc7695Fd00E3b32D08124b7a4287493aEE99f9ee"))

    @classmethod
    def for_network(cls, network: type[Network] = get_current_network()):
        if network.chain_id == Ethereum.chain_id:
            return cls.Ethereum
        if network.chain_id == Arbitrum.chain_id:
            return cls.Arbitrum



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives
from pydantic import BaseModel


class Request(BaseModel):
    params: tuple[
        primitives.address,  # tokenIn
        primitives.address,  # tokenOut
        primitives.uint24,  # fee
        primitives.address,  # recipient
        primitives.uint256,  # amountIn
        primitives.uint256,  # amountOutMinimum
        primitives.uint160,  # sqrtPriceLimitX96
    ]


class V3SwapRouter(ProtocolBase):
    exact_input_single: Annotated[
        ContractFunc[
            Request,
            primitives.uint256,
        ],
        Name("exactInputSingle"),
    ] = METHOD



from typing import Annotated

from eth_rpc import ContractFunc
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typeshed.erc20 import ERC20


class ERC4626(ERC20):
    asset: ContractFunc[
        NoArgs,
        Annotated[primitives.address, Name("assetTokenAddress")],
    ] = METHOD
    total_assets: Annotated[
        ContractFunc[
            NoArgs,
            Annotated[primitives.uint256, Name("totalManagedAssets")],
        ],
        Name("totalAssets"),
    ] = METHOD
    convert_to_shares: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("assets")],
            Annotated[primitives.uint256, Name("shares")],
        ],
        Name("convertToShares"),
    ] = METHOD
    convert_to_assets: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("shares")],
            Annotated[primitives.uint256, Name("assets")],
        ],
        Name("convertToAssets"),
    ] = METHOD
    max_deposit: Annotated[
        ContractFunc[
            Annotated[primitives.address, Name("receiver")],
            Annotated[primitives.uint256, Name("maxAssets")],
        ],
        Name("maxDeposit"),
    ] = METHOD
    preview_deposit: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("assets")],
            Annotated[primitives.uint256, Name("shares")],
        ],
        Name("previewDeposit"),
    ] = METHOD
    deposit: Annotated[
        ContractFunc[
            tuple[
                Annotated[primitives.uint256, Name("assets")],
                Annotated[primitives.address, Name("receiver")],
            ],
            Annotated[primitives.uint256, Name("shares")],
        ],
        Name("deposit"),
    ] = METHOD
    max_mint: Annotated[
        ContractFunc[
            Annotated[primitives.address, Name("receiver")],
            Annotated[primitives.uint256, Name("maxShares")],
        ],
        Name("maxMint"),
    ] = METHOD
    preview_mint: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("shares")],
            Annotated[primitives.uint256, Name("assets")],
        ],
        Name("previewMint"),
    ] = METHOD
    mint: Annotated[
        ContractFunc[
            tuple[
                Annotated[primitives.uint256, Name("shares")],
                Annotated[primitives.address, Name("receiver")],
            ],
            Annotated[primitives.uint256, Name("assets")],
        ],
        Name("mint"),
    ] = METHOD
    max_withdraw: Annotated[
        ContractFunc[
            Annotated[primitives.address, Name("owner")],
            Annotated[primitives.uint256, Name("maxAssets")],
        ],
        Name("maxWithdraw"),
    ] = METHOD
    preview_withdraw: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("assets")],
            Annotated[primitives.uint256, Name("shares")],
        ],
        Name("previewWithdraw"),
    ] = METHOD
    withdraw: Annotated[
        ContractFunc[
            tuple[
                Annotated[primitives.uint256, Name("assets")],
                Annotated[primitives.address, Name("receiver")],
                Annotated[primitives.address, Name("owner")],
            ],
            Annotated[primitives.uint256, Name("shares")],
        ],
        Name("withdraw"),
    ] = METHOD
    max_redeem: Annotated[
        ContractFunc[
            Annotated[primitives.address, Name("owner")],
            Annotated[primitives.uint256, Name("maxShares")],
        ],
        Name("maxRedeem"),
    ] = METHOD
    preview_redeem: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("shares")],
            Annotated[primitives.uint256, Name("assets")],
        ],
        Name("previewRedeem"),
    ] = METHOD
    redeem: Annotated[
        ContractFunc[
            tuple[
                Annotated[primitives.uint256, Name("shares")],
                Annotated[primitives.address, Name("receiver")],
                Annotated[primitives.address, Name("owner")],
            ],
            Annotated[primitives.uint256, Name("assets")],
        ],
        Name("redeem"),
    ] = METHOD



from .contract import ROUTER_ADDRESS, UniswapV2Router
from .types import (
    AddLiquidityETHRequest,
    AddLiquidityRequest,
    AddLiquidityResponse,
    RemoveLiquidityETHRequest,
    RemoveLiquidityETHResponse,
    RemoveLiquidityRequest,
    RemoveLiquidityResponse,
)

__all__ = [
    "ROUTER_ADDRESS",
    "UniswapV2Router",
    "AddLiquidityETHRequest",
    "AddLiquidityRequest",
    "AddLiquidityResponse",
    "RemoveLiquidityRequest",
    "RemoveLiquidityResponse",
    "RemoveLiquidityETHRequest",
    "RemoveLiquidityETHResponse",
]



from eth_rpc.types import primitives

from ..constants import Q128, Q256

ZERO = 0


def sub_in_256(x, y):
    difference = x - y
    if difference < ZERO:
        return Q256 + difference
    else:
        return difference


def get_fees(
    fee_growth_global0: primitives.uint256,
    fee_growth_global1: primitives.uint256,
    fee_growth0_low: primitives.uint256,
    fee_growth0_hi: primitives.uint256,
    fee_growth_inside0: primitives.uint256,
    fee_growth1_low: primitives.uint256,
    fee_growth1_hi: primitives.uint256,
    fee_growth_inside1: primitives.uint256,
    liquidity: primitives.uint128,
    tick_lower: primitives.int24,
    tick_upper: primitives.int24,
    tick_current: primitives.int24,
):
    fee_growth_global_0 = fee_growth_global0
    fee_growth_global_1 = fee_growth_global1
    tick_lower_fee_growth_outside_0 = fee_growth0_low
    tick_lower_fee_growth_outside_1 = fee_growth1_low
    tick_upper_fee_growth_outside_0 = fee_growth0_hi
    tick_upper_fee_growth_outside_1 = fee_growth1_hi

    tick_lower_fee_growth_below_0 = ZERO
    tick_lower_fee_growth_below_1 = ZERO
    tick_upper_fee_growth_above_0 = ZERO
    tick_upper_fee_growth_above_1 = ZERO

    if tick_current >= tick_upper:
        tick_upper_fee_growth_above_0 = sub_in_256(
            fee_growth_global_0, tick_upper_fee_growth_outside_0
        )
        tick_upper_fee_growth_above_1 = sub_in_256(
            fee_growth_global_1, tick_upper_fee_growth_outside_1
        )
    else:
        tick_upper_fee_growth_above_0 = tick_upper_fee_growth_outside_0
        tick_upper_fee_growth_above_1 = tick_upper_fee_growth_outside_1

    if tick_current < tick_lower:
        tick_lower_fee_growth_below_0 = sub_in_256(
            fee_growth_global_0, tick_lower_fee_growth_outside_0
        )
        tick_lower_fee_growth_below_1 = sub_in_256(
            fee_growth_global_1, tick_lower_fee_growth_outside_1
        )
    else:
        tick_lower_fee_growth_below_0 = tick_lower_fee_growth_outside_0
        tick_lower_fee_growth_below_1 = tick_lower_fee_growth_outside_1

    fr_t1_0 = sub_in_256(
        sub_in_256(fee_growth_global_0, tick_lower_fee_growth_below_0),
        tick_upper_fee_growth_above_0,
    )
    fr_t1_1 = sub_in_256(
        sub_in_256(fee_growth_global_1, tick_lower_fee_growth_below_1),
        tick_upper_fee_growth_above_1,
    )

    fee_growth_inside_last_0 = fee_growth_inside0
    fee_growth_inside_last_1 = fee_growth_inside1

    uncollected_fees_0 = (
        liquidity * sub_in_256(fr_t1_0, fee_growth_inside_last_0)
    ) // Q128
    uncollected_fees_1 = (
        liquidity * sub_in_256(fr_t1_1, fee_growth_inside_last_1)
    ) // Q128

    return uncollected_fees_0, uncollected_fees_1



from typing import Annotated, Optional

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_rpc.utils import to_bytes32
from eth_typing import HexAddress, HexStr
from pydantic import PrivateAttr

from .constants import ADMIN_SLOT, EIP1967_IMPLEMENTATION_SLOT, OZ_IMPLEMENTATION_SLOT
from .types import (
    ApproveRequest,
    OwnerSpenderRequest,
    TransferFromRequest,
    TransferRequest,
)


class ERC20Metadata(ProtocolBase):
    _decimals: int | None = PrivateAttr(None)
    _name: str | None = PrivateAttr(None)
    _symbol: str | None = PrivateAttr(None)

    name: ContractFunc[
        NoArgs,
        Annotated[primitives.string, Name("_name")],
    ] = METHOD
    symbol: ContractFunc[
        NoArgs,
        Annotated[primitives.string, Name("_symbol")],
    ] = METHOD
    decimals: ContractFunc[
        NoArgs,
        Annotated[primitives.uint256, Name("_decimals")],
    ] = METHOD

    def get_decimals(self):
        if not self._decimals:
            self._decimals = self.decimals().get()
        return self._decimals

    def get_name(self):
        if not self._name:
            self._name = self.name().get()
        return self._name

    def get_symbol(self):
        if not self._symbol:
            self._symbol = self.symbol().get()
        return self._symbol

    async def load_async(self):
        from eth_typeshed.multicall import multicall

        if not (self._name and self._symbol and self._decimals):
            self._name, self._symbol, self._decimals = await multicall.execute(
                self.name(),
                self.symbol(),
                self.decimals(),
            )


class ERC20(ERC20Metadata):
    balance_of: Annotated[
        ContractFunc[
            Annotated[primitives.address, Name("owner")],
            Annotated[primitives.uint256, Name("amount")],
        ],
        Name("balanceOf"),
    ] = METHOD
    total_supply: Annotated[
        ContractFunc[
            NoArgs,
            Annotated[primitives.uint256, Name("_supply")],
        ],
        Name("totalSupply"),
    ] = METHOD
    transfer: ContractFunc[
        TransferRequest,
        Annotated[bool, Name("success")],
    ] = METHOD
    transfer_from: Annotated[
        ContractFunc[
            TransferFromRequest,
            Annotated[bool, Name("success")],
        ],
        Name("transferFrom"),
    ] = METHOD
    approve: ContractFunc[
        ApproveRequest,
        Annotated[primitives.bool, Name("success")],
    ] = METHOD
    allowance: ContractFunc[
        OwnerSpenderRequest,
        Annotated[primitives.uint256, Name("result")],
    ] = METHOD

    def get_allowance_slot(self, owner: str, spender: str):
        """
        Gets the allowance slot given token/owner/spender combination
        """
        response = self._get_debug_tracecall(
            self.address,
            data=f"0xdd62ed3e000000000000000000000000{owner.replace('0x', '')}000000000000000000000000{spender.replace('0x', '')}",
        ).sync
        if not response:
            return None
        try:
            storage = response[self.address.lower()]["storage"]
        except KeyError:
            return None

        for key in [
            OZ_IMPLEMENTATION_SLOT,
            ADMIN_SLOT,
            EIP1967_IMPLEMENTATION_SLOT,
        ]:
            if hex(key) in storage:
                del storage[hex(key)]
        if len(storage) != 1:
            # TOOD: this isn't really a ValueError, but we need to allow the client to handle this differently
            raise ValueError(list(storage.keys()))
        return list(storage.keys())[0]

    def get_balance_slot(self, owner: primitives.address) -> Optional[HexStr]:
        """
        Attempts to find the balance slot for an address.
        This will not work on tokens with irregular balance calculation.
        """
        balance = self.balance_of(owner).sync.call().raw
        result = self._get_debug_tracecall(
            self.address,
            data=f"0x70a08231000000000000000000000000{owner.replace('0x', '')}",
        ).sync
        storage = result[self.address.lower()]["storage"]
        for slot in storage:
            if balance in storage[slot]:
                return slot

        # Unknown slot return None
        return None

    def balance_state_diff(
        self, owner: primitives.address, balance: int
    ) -> Optional[dict]:
        slot = self.get_balance_slot(owner)
        if slot:
            return {
                self.address: {
                    "stateDiff": {
                        slot: to_bytes32(balance),
                    },
                },
            }
        return {}

    def allowance_state_diff(
        self, owner: HexAddress, spender: HexAddress, amount: int
    ) -> Optional[dict]:
        slot = self.get_allowance_slot(owner, spender)
        if slot:
            return {
                self.address: {
                    "stateDiff": {
                        slot: to_bytes32(amount),
                    }
                },
            }
        return {}

    def __repr__(self):
        return f"<ERC20 address={self.address}>"


class ERC20BytesMetadata(ProtocolBase):
    name: ContractFunc[
        NoArgs,
        Annotated[primitives.bytes32, Name("_name")],
    ] = METHOD
    symbol: ContractFunc[
        NoArgs,
        Annotated[primitives.bytes32, Name("_symbol")],
    ] = METHOD



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typing import HexAddress
from pydantic import BaseModel


class Rewards(BaseModel):
    excluded: primitives.uint256
    realized: primitives.uint256


class TokenRewards(ProtocolBase):
    rewards_deposited: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("rewardsDeposited")
    ] = METHOD
    rewards_deposited_tkn: Annotated[
        ContractFunc[HexAddress, primitives.uint256], Name("rewardsDeposited")
    ] = METHOD
    rewards_distributed: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("rewardsDistributed")
    ] = METHOD
    rewards_dep_monthly: Annotated[
        ContractFunc[primitives.uint256, primitives.uint256], Name("rewardsDepMonthly")
    ] = METHOD
    beginning_of_month: Annotated[
        ContractFunc[
            Annotated[primitives.uint256, Name("_timestamp")], primitives.uint256
        ],
        Name("beginningOfMonth"),
    ] = METHOD
    get_unpaid: Annotated[
        ContractFunc[Annotated[HexAddress, Name("_wallet")], primitives.uint256],
        Name("getUnpaid"),
    ] = METHOD
    rewards: ContractFunc[HexAddress, Rewards] = METHOD
    rewards_token: Annotated[ContractFunc[NoArgs, HexAddress], Name("rewardsToken")] = (
        METHOD
    )
    shares: ContractFunc[HexAddress, primitives.uint256] = METHOD
    total_shares: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("totalShares")
    ] = METHOD
    total_stakers: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("totalStakers")
    ] = METHOD
    tracking_token: Annotated[
        ContractFunc[NoArgs, HexAddress], Name("trackingToken")
    ] = METHOD
